/* $Workfile: Alvis4CsiBasicDemo.c $       $Revision: 5 $      $Modtime: 10/11/08 12:33p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    Alvis4CsiBasicDemo.C

    A simple example console application demonstrating the use of OTX Driver
    with the following boards:
    
      Alvis-4-CSI
      Alvis-4M-CSI

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.


    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_LOG_FILE "Alvis4CsiBasicDemo.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_SS7_FISU_FILTERING 1

#define DEMO_MAX_MULTIPLE_FRAME_READ 128



#define DEMO_EVENT_DATA_BUF_SIZE 10

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  int                    boardNo;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_nDeviceId;     // Product model
OTX_UINT32  g_nSubsystemId;  // Product Version

struct {
  OTX_HANDLE hBoard;
  OTX_HANDLE hTempSensor;
} g_hPhysDevs;



/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    // D16V86 apps cannot use event objects
    g_hNotificationEvent = (OTX_EVENT) NULL;      
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); 
  DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); 
  DemoCheckRc(result);
}



/*******************************************************************************

  @func
    Open the Physical Devices within Alvis DMP

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    DemoLogf("Opening Alvis-4-CSI board\n");
    break;
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    DemoLogf("Opening Alvis-4M-CSI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
    break;
  }

  // Open the board device
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, 0, 0, g_hEventQueue, &g_hPhysDevs.hBoard);  DemoCheckRc(result);

  // Get the Device ID of the board
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoard, OTX_ATTR_BRD_DEVICE_ID, &g_nDeviceId); 
  DemoCheckRc(result);

  // Get the version (SubSystem ID) of the board
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoard, OTX_ATTR_BRD_SUBSYSTEM_ID, &g_nSubsystemId); 
  DemoCheckRc(result);

  DemoLogf("Board Device ID: 0x%.4X, SubsystemId: 0x%.4X\n", g_nDeviceId, g_nSubsystemId);

  DemoLogf("Open the Temperature Sensor device on this board\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoard, OTX_DEVICE_TEMPERATURE_SENSOR, 0, 0, g_hEventQueue, &g_hPhysDevs.hTempSensor);  DemoCheckRc(result);

}




/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream 
    (a subset of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. 
    A pipe is a characterized by its capacity (in 8 kbit/s increments). 
    The user can create a pipe with the driver function <f OtxDrvCreatePipe>. 
    A pipe is connected to physical or logical devices and mapped to specific 
    time-slots with the driver function <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{

}




/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{

}


/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hBoard,  OTX_TASK_SYNC, 5000L); 
  DemoCheckRc(result);

  // Enable Temp Sensor
  if (g_nSubsystemId > OTX_NIC_SUBSYSTEM_ID_REV_1_0) {  // The interrupt from the Temp Sensor device is only available in versions greater than 1.0
    result = OtxDrvEnable(g_hPhysDevs.hTempSensor,  OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{

}




/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 
  // (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a 
  // very accurate way to report timestamp)
  nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                         (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}


/*******************************************************************************
  @func

*******************************************************************************/
void  *DemoTimeStampTest(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  for (;;) {
    
    result = OtxDrvGetCurrentTime(&nTimeStamp); 
    DemoCheckRc(result);
        
    // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
    #if defined(__linux__)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
    #elif defined(WIN32)
      nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
    #endif
    
    // Convert to seconds since Jan 1, 1970 
    // (so we can use this time with OsaStdLocalTime())
    nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
    
    // Calculate number of milliseconds (since number of seconds 
    // is not a very accurate way to report timestamp)
    nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                           (nTimeStampSec.nUint64 * 1000);
    
    nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
    ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
    
    if (ptCurrentDayAndTime) {
#if !defined(_WIN32_WCE)
      DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
        ptCurrentDayAndTime->tm_hour,
        ptCurrentDayAndTime->tm_min, 
        ptCurrentDayAndTime->tm_sec,
        nTimeStampMs.sUint32.nLow,
        asctime((struct tm *) ptCurrentDayAndTime)  
        );
#endif  // #if !defined(_WIN32_WCE)
    } else {
      DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
    }

    OsaThreadSleep(4000);
  }
  


  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {

      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, 
                                    sqEventData, 
                                    DEMO_EVENT_DATA_BUF_SIZE, 
                                    &coEventsFetched); 
      DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    {
      switch (pEventData->m_nCode) {
      case OTX_BRD_EC_CORE2CORE_INTERRUPT:
        {
          DemoLogf("Core-to-Core interrupt\n");  // OTX Internal use
        }
        break;
      default:
        DemoLogf("DemoPrintEvent eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
          pEventData->m_eDeviceType,
          pEventData->m_nSourceId,
          pEventData->m_nCode);
        break;
      }
    }
    break;

  case OTX_DEVICE_TEMPERATURE_SENSOR:
    {
      switch (pEventData->m_nCode) {
      case OTX_TEMP_EC_LIMIT_EXCEEDED:  // This interrupt is only available for Alvis-CSI versions greater than 1.0 (OTX_NIC_SUBSYSTEM_ID_REV_1_0)
        {
          OTX_INT8 currentDegC, maxDegC, limitDegC;
          OTX_UINT32 nTemp;
          OTX_RESULT result;
          
          DemoLogf("Board Temperature Limit exceeded. ");
          
          result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hTempSensor, OTX_ATTR_TEMP_TEMPERATURE_VALUES, &nTemp); 
          
          if (OTX_SUCCESS(result)) {
            currentDegC = (OTX_INT8) (nTemp >> 0)  & 0xFF;
            limitDegC   = (OTX_INT8) (nTemp >> 8)  & 0xFF;
            maxDegC     = (OTX_INT8) (nTemp >> 16) & 0xFF;
            DemoLogf("Board Temp = %dC (Limit=%dC, Max=%dC)", currentDegC, limitDegC, maxDegC);
          }
          DemoLogf("\n");
        }
        break;
      default:
        DemoLogf("DemoPrintEvent eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
          pEventData->m_eDeviceType,
          pEventData->m_nSourceId,
          pEventData->m_nCode);
        break;
      }
    }
    break;

  default:
    DemoLogf("DemoPrintEvent eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
      pEventData->m_eDeviceType,
      pEventData->m_nSourceId,
      pEventData->m_nCode);
    break;
  }

}


/*******************************************************************************

  @func

*******************************************************************************/
void DemoCloseDevices(
  void
)
{


}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_DMP_ALVIS_4M_CSI;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if (*(argv[i] + 2) == 'D') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_CSI;
          } else if (*(argv[i] + 2) == 'K') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4M_CSI;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[D|K]        : Board Type used for the demo, where\n");
  DemoLogf("                    D  = Use Alvis-4-CSI board\n");
  DemoLogf("                    K  = Use Alvis-4M-CSI board\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile, formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", 
      OTX_RESULT_CODE(result), 
      OTX_RESULT_SOURCE_ID(result), 
      OTX_RESULT_LINE_NUM(result), 
      msgStr);

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" g   Get current time\n");
    DemoLogf(" t   Get board temperature\n");
    DemoLogf(" T   Reset Max temperature value\n");
    DemoLogf(" ^   Set temperature limit\n");
    DemoLogf(" r   Read 32 bit register\n");
    DemoLogf(" R   Read a block of 32 bit register\n");
    DemoLogf(" w   Write 32 bit register\n");
    DemoLogf("===  ======== [The following commands are only used for Odin internal use]\n");
    DemoLogf(" d   Send DMP Debug Command (generic)\n");
    DemoLogf(" D   Send Alvis-4-CSI/Alvis-4M-CSI Debug Command (board specific)\n");
    break;


#if 1 // For OTX internal debugging
  case 'd':
    {
      OTX_UINT32 nCmd;

      DemoLogf("DMP Debug Command. Select Command:\n");
      DemoLogf("0 = Read GPIO registers\n");
      DemoLogf("1 = Issue Core interrupt\n");
      nCmd = OsaConsoleGetCh() - '0';

      DemoLogf("Command value 0x%X sent\n", nCmd);

      OtxDrvDbgDump(g_hPhysDevs.hBoard, nCmd);

      DemoLogf("Done.\n");
    }
    break;
  case 'D':
    {
      OTX_UINT32 nCmd;

      DemoLogf("Alvis-4-CSI Debug Command. Select Command:\n");
      DemoLogf("0 = Simulate T1E1 device interrupt\n");
      DemoLogf("1 = Read RTC FW version\n");
      DemoLogf("2 = Read Board Temperature\n");
      DemoLogf("3 = Reset Min/Max Temperature\n");
      DemoLogf("4 = Set TP10 high\n");
      DemoLogf("5 = Set TP10 low\n");
      DemoLogf("6 = Read RTC ports (2 and 3)\n");
      DemoLogf("7 = Read Eeprom (Address 0x7F00)\n");
      DemoLogf("9 = Read Eeprom (Address 0x0000)\n");

      nCmd = OsaConsoleGetCh() - '0' + 0x1000;

      DemoLogf("Command value 0x%X sent\n", nCmd);

      OtxDrvDbgDump(g_hPhysDevs.hBoard, nCmd);

      DemoLogf("Done.\n");
    }
    break;
#endif // For OTX internal debugging

  case 'g':
    {
      OTX_DATETIME nTimeStamp;
      OTX_DATETIME nTimeStampMs;
      OTX_DATETIME nTimeStampSec;
      struct Osa_tm *ptCurrentDayAndTime;
      Osa_time_t nCurrentTimeSecs;
      
      result = OtxDrvGetCurrentTime(&nTimeStamp); 
      DemoCheckRc(result);
      
      DemoLogf("%.8X%.8X (Current time)\n", 
               nTimeStamp.sUint32.nHigh, 
               nTimeStamp.sUint32.nLow);
      
      // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
      #if defined(__linux__)
        nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
      #elif defined(WIN32)
        nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
      #endif
      
      // Convert to seconds since Jan 1, 1970 
      // (so we can use this time with OsaStdLocalTime())
      nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
      
      // Calculate number of milliseconds (since number of seconds 
      // is not a very accurate way to report timestamp)
      nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                             (nTimeStampSec.nUint64 * 1000);
      
      nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
      ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
      
      if (ptCurrentDayAndTime) {
#if !defined(_WIN32_WCE)
        DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
          ptCurrentDayAndTime->tm_hour,
          ptCurrentDayAndTime->tm_min, 
          ptCurrentDayAndTime->tm_sec,
          nTimeStampMs.sUint32.nLow,
          asctime((struct tm *) ptCurrentDayAndTime)
          );
#endif  // #if !defined(_WIN32_WCE)
      } else {
        DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
      }
    }
    break;


  case 't':
    {
      OTX_UINT32 nTemp;
      OTX_INT8 currentDegC, maxDegC, limitDegC;

      DemoLogf("Get board temperature\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hTempSensor, OTX_ATTR_TEMP_TEMPERATURE_VALUES, &nTemp); 
      DemoCheckRc(result);
      
      currentDegC = (OTX_INT8) (nTemp >> 0)  & 0xFF;
      limitDegC   = (OTX_INT8) (nTemp >> 8)  & 0xFF;
      maxDegC     = (OTX_INT8) (nTemp >> 16) & 0xFF;
      //DemoLogf("nTemp=0x%.8X\n", nTemp);
      DemoLogf("Board Temp = %dC  (Limit=%dC, Max=%dC)\n", currentDegC, limitDegC, maxDegC);
    }
    break;

  case 'T':
    {
      DemoLogf("Reset Min/Max temperature values\n");

      //result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTempSensor, OTX_ATTR_TEMP_TEMPERATURE_VALUES, 0); 
      result = OtxDrvReset(g_hPhysDevs.hTempSensor); 
      DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case '^':
    {
      static OTX_UINT32 nLimitTemp = 75;
      char strBuf[12];

      DemoLogf("Set temperature limit\n");

      DemoLogf("Enter value between 1 and 127 (%d): ", nLimitTemp);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nLimitTemp = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTempSensor, OTX_ATTR_TEMP_TEMPERATURE_VALUES, nLimitTemp); 
      DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;


  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 32-bit register \n");
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg32(g_hPhysDevs.hBoard, nAddress, &readVal); 
      DemoCheckRc(result);

      DemoLogf("Address 0x%X = 0x%.8X\n", nAddress, readVal);
    }
    break;

  case 'R':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      OTX_UINT32 nFormatCnt      = 0;
      OTX_UINT32 nLocAddr;
      OTX_UINT32 nLocCount;
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nCount   = 0;

      DemoLogf("Read from multiple 32-bit register \n");
      DemoLogf("Starting Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      DemoLogf("Count (0x%X): ", nCount);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nCount = strtoul(strBuf, NULL, 0);
      }

      nLocAddr = nAddress;
      nLocCount = nCount;
      if(nLocCount) {
        while(nLocCount) {
          if(!(nFormatCnt % 4)) {
            DemoLogf("\n");
            DemoLogf("Address 0x%.8X ", nLocAddr);
          } 
          result = OtxDrvReadReg32(g_hPhysDevs.hBoard, nLocAddr, &readVal);
          DemoCheckRc(result);

          DemoLogf(" 0x%.8X  ", readVal);

          nFormatCnt++;
          nLocCount--;
          nLocAddr = nLocAddr + 4;
        }

        DemoLogf("\n");
      }
      else {
        DemoLogf("Count set to zero should be set to 1 or more\n");
      }

      DemoLogf("\n");
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 32-bit register \n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg32(g_hPhysDevs.hBoard, nAddress, nValue); 
      DemoCheckRc(result);

      DemoLogf("Wrote 0x%.8X Address 0x%X\n", nValue, nAddress);
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("Alvis4CsiBasicDemo, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Device Example program.\n");
  DemoLogf("For Alvis-4-CSI and Alvis-4M-CSI\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_REV_STR, 
                                       szVersionInfo, 
                                       255); 
    DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, 
                                         OTX_ATTR_LIB_TYPE_STATIC, 
                                         &bIsStaticLinkedLibrary); 
    DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_FILE_NAME, 
                                       szLibFileName, 
                                       64); 
    DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", 
             szVersionInfo, 
             szLibFileName, 
             bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, 
                                       OTX_ATTR_DRV_REV_STR, 
                                       szVersionInfo, 
                                       255); 
    DemoCheckRc(result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");


  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  /*
  result = OtxBrdSetClocks(g_hPhysDevs.hBoard, 
                           OTX_CLOCK_MODE_LOCAL, 
                           OTX_CLOCK_SOURCE_LOCAL_0); 
  DemoCheckRc(result);
  */

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_UINT32      cbAttribValue;  
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoard, 
                                    OTX_ATTR_BRD_SERIAL_NUMBER, 
                                    &eAttribType, 
                                    &cbAttribValue, 
                                    &bReadOnly); 
    DemoCheckRc(result);

    if(!(eAttribType == OTX_ATTRIB_SZ)) {
      DemoLogf("Warning: eAttribType should equal OTX_ATTRIB_SZ\n");
    }

    if(!(bReadOnly == OTX_TRUE)) {
      DemoLogf("Warning: bReadOnly should equal OTX_TRUE\n");
    }

    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoard, 
                                       OTX_ATTR_BRD_SERIAL_NUMBER, 
                                       pszSerialNumber, 
                                       cbAttribValue); 
    DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_UINT32      cbAttribValue;  
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoard, 
                                    OTX_ATTR_BRD_BUILD, 
                                    &eAttribType, 
                                    &cbAttribValue, 
                                    &bReadOnly); 
    DemoCheckRc(result);

    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoard, 
                                       OTX_ATTR_BRD_BUILD, 
                                       pszBrdBuildNumber, 
                                       cbAttribValue); 
    DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read FW Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszFwNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoard, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszFwNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoard, OTX_ATTR_BRD_CPLD_BUILD, pszFwNumbers, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Firmware:\n%s", pszFwNumbers);
    free(pszFwNumbers);
  }



  // Main Loop. Loop here and wait for 
  // Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  DemoPrintf("Exit Alvis-CSI Demo\n");

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}


